function [Cxy_CWT,f,t,pairs]=ISC_CWT(IMF,fs,fr)
% Copyrighted to Apit Hemakom and Danilo P. Mandic 2017

% ISC_CWT function calculates time-varying coherence, which takes into account both amplitude and phase information, based on highly localised univariate and bivariate time-frequency representations
% generated using CWT-based SST and MSST of IMFs obtained from NA-MEMD.
% When performing ISC algorithm, 30 realisations of NA-MEMD should be performed to reduce the leakage effect in the input channels, which is caused by artificial WGN
% channels while performing NA-MEMD
% Also, for NA-MEMD a number of adjacent WGN channels, at least 10, shall be used to ensure that the number of IMFs from different realisations of NA-MEMD is highly consistent.

% ---- Input ----
% IMF: M x N x O information-bearing IMFs obtained from NA-MEMD, where M is the number of channels, N the number of IMFs, and O the number of data points.
% fs: Sampling frequency.
% fr: Frequency resolution.

% ---- Output ----
% Cxy_CWT: Time-varying coherence with the dimensions of N x P x Q x R, where N is the number of IMFs, P the number of channel pairs determinded using P = nchoosek(M,2), Q frequency bins, and R the number of time windows.
% f: Frequency bins.
% t: Time windows.
% pairs: P x 2 matrix showing the order of channel pairs corresponding to P channel pairs.

% Note - 'single' conversion is used in order to save memory space, which otherwise would be largely consumed by the standard 'double' precision.
% From MathWorks, "Because MATLAB stores numbers of type single using 32 bits, they require less memory than numbers of type double , which use 64 bits."

% 'multi_sst_TF_main' function from A. Ahrabian and D. P. Mandic, "Selective time-frequency reassignment based on synchrosqueezing," IEEE Signal Processing Letters, vol. 22, no. 11, pp. 2039-2043, 2015. is required and 
% and can be downloaded from http://www.commsp.ee.ic.ac.uk/~mandic/research/Multivariate_SST_code.zip
%%
t=(1:length(IMF))/fs;

NUM_CHs=size(IMF,1);
NUM_IMFs=size(IMF,2);

% all possible pairs
pairs=nchoosek(1:NUM_CHs,2);

% total number of possible pairs
NUM_PAIRS=size(pairs,1);

x(:,1)=squeeze(IMF(pairs(1,1),1,:));
x(:,2)=squeeze(IMF(pairs(1,2),1,:));
[Tf,~,sst]=multi_sst_TF_main_v2(x,5,1,fs,fr);
Cxy_CWT=single(zeros(NUM_IMFs,NUM_PAIRS,size(Tf,1),size(Tf,2)));
        
for pair_count=1:NUM_PAIRS
    for i=1:NUM_IMFs

        x(:,1)=squeeze(IMF(pairs(pair_count,1),i,:));
        x(:,2)=squeeze(IMF(pairs(pair_count,2),i,:));
        [Tf,~,sst]=multi_sst_TF_main_v2(x,5,0,fs,fr); % 1 is worse, more ripples
        f=linspace(0,fs/2,size(sst,2));
        
        S12_temp=single(Tf);
        
        Tf=[];
        
        S1_temp=single(squeeze(sst(1,:,:)));
        S2_temp=single(squeeze(sst(2,:,:)));
        
        sst=[]; 

        S12=single(sqrt(S12_temp));
        
        S12_temp=[];

        S1=single(abs(S1_temp));
        
        S1_temp=[];
        
        S2=single(abs(S2_temp));
                
        S2_temp=[];
        
        S_1_2=single(sqrt(S1.*S2));
        
        S1=[];
        S2=[];

        Cxy_CWT_temp = single(S_1_2./S12);

        S_1_2=[]; 
        S12=[]; 

        Cxy_CWT_temp(find(Cxy_CWT_temp==Inf))=0;
        Cxy_CWT_temp(isnan(Cxy_CWT_temp))=0;

        Cxy_CWT_temp=single(Cxy_CWT_temp/max(max(Cxy_CWT_temp)));
                
        Cxy_CWT(i,pair_count,:,:) = Cxy_CWT_temp;
        
        Cxy_CWT_temp=[];

%         if pair_count==1
%             S_sst(1,:,:)=single(S1);
%             S_sst(2,:,:)=single(S2);
%         elseif pair_count==2
%             S_sst(3,:,:)=single(S2);
%         elseif pair_count==3
%             S_sst(4,:,:)=single(S2);    
%         elseif pair_count==4
%             S_sst(5,:,:)=single(S2);
%         elseif pair_count==5
%             S_sst(6,:,:)=single(S2);
%         end
        
%         S_msst(pair_count,:,:)=single(S12);
                
    end
end

