function [Cxy_chunk_Theta,Cxy_chunk_Alpha,Cxy_chunk_Beta,f_Theta,f_Alpha,f_Beta,t,pairs]=memd_msst_coh_stft_based_v2_3_EEG(IMF,fs,fr,Th_std)
% Copyrighted to Apit Hemakom and Danilo P. Mandic 2017
% Change in ver.2
% - enabling user-specified frequency resolution, fr, for synchrosqueezing
% Change in ver.2.1
% - S_sst & S_msst output variables are no longer available to save memory
% - optimised memory usage
% Change in ver.2.1.EEG
% - coherence value in each EEG band is calculated separately
% Change in ver.2.2.EEG
% - coherence value in only Theta, Alpha and Beta bands are calculated
% Change in ver.2.3.EEG
% - outliers outside Th_std*std in each EEG band are limited to Th_std*std

% MEMD_MSST_COH function calculates time-varying coherence, which takes into account both amplitude and phase information, based on highly localised univariate and bivariate time-frequency representations
% generated using SST and MSST of IMFs obtained from NA-MEMD.
% When performing MEMD-MSST-COH algorithm, a number of realisations of NA-MEMD must be performed, at least 30 is recommended, to be able to average out mathematical artefacts (noise) generated from SST/MSST. 
% Also, a number of adjacent WGN channels, at least 10, shall be used to ensure that the number of IMFs from different realisations of NA-MEMD is consistent.

% ---- Input ----
% IMF: M x N x O information-bearing IMFs obtained from NA-MEMD, where M is the number of channels, N the number of IMFs, and O the number of data points.
% fs: Sampling frequency.
% ---- Output ----
% Cxy_chunk: Time-varying coherence with the dimensions of N x P x Q x R, where N is the number of IMFs, P the number of channel pairs determinded using P = nchoosek(M,2), Q frequency bins, and R the number of time windows.
% f: Frequency bins.
% t: Time windows.
% S_sst: M x Q x R univariate TF representation for each channel generated using SST.
% S_msst: P x Q x R bivariate TF representation for each channel pair generated using MSST.
% S_sst and S_msst are no longer available in ver.2.1 in order to save
% memory.

% pairs: P x 2 matrix showing the order of channel pairs corresponding to P channel pairs.

% Note - 'single' conversion is used in order to save memory space, which otherwise would be largely consumed by the standard 'double' precision.
% From MathWorks, "Because MATLAB stores numbers of type single using 32 bits, they require less memory than numbers of type double , which use 64 bits."

% 'multi_sst_TF_main' function from A. Ahrabian and D. P. Mandic, "Selective time-frequency reassignment based on synchrosqueezing," IEEE Signal Processing Letters, vol. 22, no. 11, pp. 2039-2043, 2015. is required and 
% and can be downloaded from http://www.commsp.ee.ic.ac.uk/~mandic/research/Multivariate_SST_code.zip
%%
t=(1:length(IMF))/fs;

NUM_CHs=size(IMF,1);
NUM_IMFs=size(IMF,2);

% all possible pairs
pairs=nchoosek(1:NUM_CHs,2);

% total number of possible pairs
NUM_PAIRS=size(pairs,1);

% x(:,1)=squeeze(IMF(pairs(1,1),1,:));
% x(:,2)=squeeze(IMF(pairs(1,2),1,:));


% [Tf,~,sst]=multi_sst_TF_main_stft_based_v2(x,5,1,fs,fr);

%  f=linspace(0,fs/2,size(sst,2));
 f=linspace(0,fs/2,(fs/2)/fr);
% Delta_lower_f=0.0001;
% Delta_upper_f=2.9999;
 
Theta_lower_f=3;
Theta_upper_f=6.9999;

Alpha_lower_f=7;
Alpha_upper_f=12.9999;

Beta_lower_f=13;
Beta_upper_f=29.9999;

% Gamma_lower_f=30;
% Gamma_upper_f=40.9999;

% [~,Delta_lower_f_index]=min(abs(f-Delta_lower_f));
% [~,Delta_upper_f_index]=min(abs(f-Delta_upper_f));
[~,Theta_lower_f_index]=min(abs(f-Theta_lower_f));
[~,Theta_upper_f_index]=min(abs(f-Theta_upper_f));
[~,Alpha_lower_f_index]=min(abs(f-Alpha_lower_f));
[~,Alpha_upper_f_index]=min(abs(f-Alpha_upper_f));
[~,Beta_lower_f_index]=min(abs(f-Beta_lower_f));
[~,Beta_upper_f_index]=min(abs(f-Beta_upper_f));
% [~,Gamma_lower_f_index]=min(abs(f-Gamma_lower_f));
% [~,Gamma_upper_f_index]=min(abs(f-Gamma_upper_f));

% f_Delta=f(Delta_lower_f_index:Delta_upper_f_index);
f_Theta=f(Theta_lower_f_index:Theta_upper_f_index);
f_Alpha=f(Alpha_lower_f_index:Alpha_upper_f_index);
f_Beta=f(Beta_lower_f_index:Beta_upper_f_index);
% f_Gamma=f(Gamma_lower_f_index:Gamma_upper_f_index);

% Cxy_chunk_Delta=single(zeros(NUM_IMFs,NUM_PAIRS,Delta_upper_f_index-Delta_lower_f_index+1,size(Tf,2)));
Cxy_chunk_Theta=single(zeros(NUM_IMFs,NUM_PAIRS,Theta_upper_f_index-Theta_lower_f_index+1,size(IMF,3)));
Cxy_chunk_Alpha=single(zeros(NUM_IMFs,NUM_PAIRS,Alpha_upper_f_index-Alpha_lower_f_index+1,size(IMF,3)));
Cxy_chunk_Beta=single(zeros(NUM_IMFs,NUM_PAIRS,Beta_upper_f_index-Beta_lower_f_index+1,size(IMF,3)));
% Cxy_chunk_Gamma=single(zeros(NUM_IMFs,NUM_PAIRS,Gamma_upper_f_index-Gamma_lower_f_index+1,size(Tf,2)));
% Tf=[];
% x=[];
% sst=[];
for pair_count=1:NUM_PAIRS
    for i=1:NUM_IMFs
        x=[];
        x(:,1)=squeeze(IMF(pairs(pair_count,1),i,:));
        x(:,2)=squeeze(IMF(pairs(pair_count,2),i,:));
                 
        [Tf,~,sst]=multi_sst_TF_main_stft_based_v2(x,5,0,fs,fr); % 1 is worse, more ripples
     
        S12=single(sqrt(Tf));
        
        Tf=[];

        S1=single(abs(squeeze(sst(1,:,:))));
        S2=single(abs(squeeze(sst(2,:,:))));
                
        sst=[];
        
        S_1_2=single(sqrt(S1.*S2));
        
        S1=[];
        S2=[];

        msst_coh = single(S_1_2./S12);

        S_1_2=[];
        S12=[];
        
        msst_coh(find(msst_coh==Inf))=0;
        msst_coh(isnan(msst_coh))=0;

        %%
        msst_coh_Theta=msst_coh(Theta_lower_f_index:Theta_upper_f_index,:);
        msst_coh_Alpha=msst_coh(Alpha_lower_f_index:Alpha_upper_f_index,:);
        msst_coh_Beta=msst_coh(Beta_lower_f_index:Beta_upper_f_index,:);

        msst_coh=[];
        %% Removing 'outliers' in Theta band, and calculate MSST-COH
        
        Theta_max=max(msst_coh_Theta);%vector
        Theta_max_mean=mean(Theta_max);%scalar
        Theta_max_std=std(Theta_max);%scalar

        I_temp_Theta=find(msst_coh_Theta>Th_std*Theta_max_std);
       
        msst_coh_Theta(I_temp_Theta)=Th_std*Theta_max_std;
        
        Cxy_chunk_Theta(i,pair_count,:,:)=single(msst_coh_Theta/max(max(msst_coh_Theta)));
        
        msst_coh_Theta=[];
        %% Removing 'outliers' in Alpha band, and calculate MSST-COH
         
        Alpha_max=max(msst_coh_Alpha);%vector
        Alpha_max_mean=mean(Alpha_max);%scalar
        Alpha_max_std=std(Alpha_max);%scalar

        I_temp_Alpha=find(msst_coh_Alpha>Th_std*Alpha_max_std);
       
        msst_coh_Alpha(I_temp_Alpha)=Th_std*Alpha_max_std;
        
         Cxy_chunk_Alpha(i,pair_count,:,:) =single(msst_coh_Alpha/max(max(msst_coh_Alpha)));
         
         msst_coh_Alpha=[];
        %% Removing 'outliers' in Beta band, and calculate MSST-COH
                
        Beta_max=max(msst_coh_Beta);%vector
        Beta_max_mean=mean(Beta_max);%scalar
        Beta_max_std=std(Beta_max);%scalar

        I_temp_Beta=find(msst_coh_Beta>Th_std*Beta_max_std);
       
        msst_coh_Beta(I_temp_Beta)=Th_std*Beta_max_std;
        
        Cxy_chunk_Beta(i,pair_count,:,:)=single(msst_coh_Beta/max(max(msst_coh_Beta)));
        
        msst_coh_Beta=[];
                        
    end
end

