function [Cxy_STFT,f,t,pairs]=ISC_STFT(IMF,fs,fr,Th_std,W_STFT,alpha)
% Copyrighted to Apit Hemakom and Danilo P. Mandic 2017

% ISC_STFT function calculates time-varying coherence, which takes into account both amplitude and phase information, based on highly localised univariate and bivariate time-frequency representations
% generated using STFT-based SST and MSST of IMFs obtained from NA-MEMD.
% When performing ISC algorithm, 30 realisations of NA-MEMD should be performed to reduce the leakage effect in the input channels, which is caused by artificial WGN
% channels while performing NA-MEMD
% Also, for NA-MEMD a number of adjacent WGN channels, at least 10, shall be used to ensure that the number of IMFs from different realisations of NA-MEMD is highly consistent.

% ---- Input ----
% IMF: M x N x O information-bearing IMFs obtained from NA-MEMD, where M is the number of channels, N the number of IMFs, and O the number of data points.
% fs: Sampling frequency.
% fr: Frequency resolution.
% Th_std: Threshold determined by standard deviation.
% W_STFT: Windows size of STFT.
% alpha: the reciprocal of the standard deviation and is a measure of the width of its Fourier Transform.
%   As ALPHA increases, the width of the window will decrease. A value of 2.5 is recommended. See 'gausswin' function.

% ---- Output ----
% Cxy_STFT: Time-varying coherence with the dimensions of N x P x Q x R, where N is the number of IMFs, P the number of channel pairs determinded using P = nchoosek(M,2), Q frequency bins, and R the number of time windows.
% f: Frequency bins.
% t: Time windows.
% pairs: P x 2 matrix showing the order of channel pairs corresponding to P channel pairs.

% Note - 'single' conversion is used in order to save memory space, which otherwise would be largely consumed by the standard 'double' precision.
% From MathWorks, "Because MATLAB stores numbers of type single using 32 bits, they require less memory than numbers of type double , which use 64 bits."

% 'multi_sst_TF_main' function from A. Ahrabian and D. P. Mandic, "Selective time-frequency reassignment based on synchrosqueezing," IEEE Signal Processing Letters, vol. 22, no. 11, pp. 2039-2043, 2015. is required and 
% and can be downloaded from http://www.commsp.ee.ic.ac.uk/~mandic/research/Multivariate_SST_code.zip
%%
t=(1:length(IMF))/fs;

NUM_CHs=size(IMF,1);
NUM_IMFs=size(IMF,2);

% all possible pairs
pairs=nchoosek(1:NUM_CHs,2);

% total number of possible pairs
NUM_PAIRS=size(pairs,1);

x(:,1)=squeeze(IMF(pairs(1,1),1,:));
x(:,2)=squeeze(IMF(pairs(1,2),1,:));
[Tf,~,~]=multi_sst_TF_main_stft_based_v2(x,5,1,fs,fr,W_STFT,alpha);
Cxy_STFT=single(zeros(NUM_IMFs,NUM_PAIRS,size(Tf,1),size(Tf,2)));
Tf=[];
for pair_count=1:NUM_PAIRS
    for i=1:NUM_IMFs

        x(:,1)=squeeze(IMF(pairs(pair_count,1),i,:));
        x(:,2)=squeeze(IMF(pairs(pair_count,2),i,:));
        [Tf,~,sst]=multi_sst_TF_main_stft_based_v2(x,5,0,fs,fr,W_STFT,alpha); % 1 is worse, more ripples
                       
        f=linspace(0,fs/2,size(sst,2));
        
        S12_temp=single(Tf); % MSST coefficients of channels A & B
        
        Tf=[]; % clear variable to save memory
        
        S1_temp=single(squeeze(sst(1,:,:))); % SST coefficients of channel A
        S2_temp=single(squeeze(sst(2,:,:))); % SST coefficients of channel B
        
        sst=[]; % clear variable to save memory

        S12=single(sqrt(S12_temp));
        
        S12_temp=[]; % clear variable to save memory

        S1=single(abs(S1_temp));
        
        S1_temp=[]; % clear variable to save memory
        
        S2=single(abs(S2_temp));
        
        S2_temp=[]; % clear variable to save memory
        
        S_1_2=single(sqrt(S1.*S2));
        
        S1=[]; % clear variable to save memory
        S2=[]; % clear variable to save memory

        Cxy_STFT_temp = single(S_1_2./S12);
        
        S_1_2=[]; % clear variable to save memory
        S12=[]; % clear variable to save memory

        Cxy_STFT_temp(find(Cxy_STFT_temp==Inf))=0; % If the resulting division is not a number (infinitiny) due to a value of zero of the denominator, which 
                                                                        % indicates the non-existence of smearing, it must be replaced by zero.
        Cxy_STFT_temp(isnan(Cxy_STFT_temp))=0;

        %% Limit outliers with values larger than 'Th_std*std' to 'Th_std*std'
        
        Cxy_temp_max=max(Cxy_STFT_temp); % maximum value in each row, i.e. time
                
        msst_coh_max_std=std(Cxy_temp_max); 
        
        Cxy_temp_max=[]; % clear variable to save memory

        I=find(Cxy_STFT_temp>Th_std*msst_coh_max_std);
       
        Cxy_STFT_temp(I)=Th_std*msst_coh_max_std;
                
        %%
                
        Cxy_STFT_temp=single(Cxy_STFT_temp/max(max(Cxy_STFT_temp)));
                
        Cxy_STFT(i,pair_count,:,:) = single(Cxy_STFT_temp);

        Cxy_STFT_temp=[]; % clear variable to save memory
                
    end
end

