function [Cxy_Delta,Cxy_Theta,Cxy_Alpha,Cxy_Beta,Cxy_Gamma,f_Delta,f_Theta,f_Alpha,f_Beta,f_Gamma,t,pairs]=ISC_STFT_EEG(IMF,fs,fr,Th_std,W_STFT,alpha,Freq_band)
% Copyrighted to Apit Hemakom and Danilo P. Mandic 2017

% ISC_STFT_EEG function calculates time-varying coherence in EEG signals, which takes into account both amplitude and phase information, based on highly localised univariate and bivariate time-frequency representations
% generated using STFT-based SST and MSST of IMFs obtained from NA-MEMD.
% When performing ISC algorithm, 30 realisations of NA-MEMD should be performed to reduce the leakage effect in the input channels, which is caused by artificial WGN
% channels while performing NA-MEMD
% Also, for NA-MEMD a number of adjacent WGN channels, at least 10, shall be used to ensure that the number of IMFs from different realisations of NA-MEMD is highly consistent.

% EEG frequency bands
% Delta: 1-3 Hz
% Theta: 3-7 Hz
% Alpha: 7-13 Hz
% Beta: 13-30 Hz
% Gamma: 30-50 Hz

% ---- Input ----
% IMF: M x N x O information-bearing IMFs obtained from NA-MEMD, where M is the number of channels, N the number of IMFs, and O the number of data points.
% fs: Sampling frequency.
% fr: Frequency resolution.
% Th_std: Threshold determined by standard deviation.
% W_STFT: Windows size of STFT.
% alpha: the reciprocal of the standard deviation and is a measure of the width of its Fourier Transform.
%   As ALPHA increases, the width of the window will decrease. A value of 2.5 is recommended. See 'gausswin' function.

% ---- Output ----
% Cxy_STFT: Time-varying coherence with the dimensions of N x P x Q x R, where N is the number of IMFs, P the number of channel pairs determinded using P = nchoosek(M,2), Q frequency bins, and R the number of time windows.
% f: Frequency bins.
% t: Time windows.
% pairs: P x 2 matrix showing the order of channel pairs corresponding to P channel pairs.

% Note - 'single' conversion is used in order to save memory space, which otherwise would be largely consumed by the standard 'double' precision.
% From MathWorks, "Because MATLAB stores numbers of type single using 32 bits, they require less memory than numbers of type double , which use 64 bits."

% 'multi_sst_TF_main' function from A. Ahrabian and D. P. Mandic, "Selective time-frequency reassignment based on synchrosqueezing," IEEE Signal Processing Letters, vol. 22, no. 11, pp. 2039-2043, 2015. is required and 
% and can be downloaded from http://www.commsp.ee.ic.ac.uk/~mandic/research/Multivariate_SST_code.zip
%%

Delta_lower_f=Freq_band(1,1);
Delta_upper_f=Freq_band(2,1);

Theta_lower_f=Freq_band(1,2);
Theta_upper_f=Freq_band(2,2);

Alpha_lower_f=Freq_band(1,3);
Alpha_upper_f=Freq_band(2,3);

Beta_lower_f=Freq_band(1,4);
Beta_upper_f=Freq_band(2,4);

Gamma_lower_f=Freq_band(1,5);
Gamma_upper_f=Freq_band(2,5);

t=(1:length(IMF))/fs;

NUM_CHs=size(IMF,1);
NUM_IMFs=size(IMF,2);

% all possible pairs
pairs=nchoosek(1:NUM_CHs,2);

% total number of possible pairs
NUM_PAIRS=size(pairs,1);

 f=linspace(0,fs/2,(fs/2)/fr);
% Delta_lower_f=1;
% Delta_upper_f=2.9999;
%  
% Theta_lower_f=3;
% Theta_upper_f=6.9999;
% 
% Alpha_lower_f=7;
% Alpha_upper_f=12.9999;
% 
% Beta_lower_f=13;
% Beta_upper_f=29.9999;

% Gamma_lower_f=30;
% Gamma_upper_f=40.9999;

[~,Delta_lower_f_index]=min(abs(f-Delta_lower_f));
[~,Delta_upper_f_index]=min(abs(f-Delta_upper_f));
[~,Theta_lower_f_index]=min(abs(f-Theta_lower_f));
[~,Theta_upper_f_index]=min(abs(f-Theta_upper_f));
[~,Alpha_lower_f_index]=min(abs(f-Alpha_lower_f));
[~,Alpha_upper_f_index]=min(abs(f-Alpha_upper_f));
[~,Beta_lower_f_index]=min(abs(f-Beta_lower_f));
[~,Beta_upper_f_index]=min(abs(f-Beta_upper_f));
[~,Gamma_lower_f_index]=min(abs(f-Gamma_lower_f));
[~,Gamma_upper_f_index]=min(abs(f-Gamma_upper_f));

f_Delta=f(Delta_lower_f_index:Delta_upper_f_index);
f_Theta=f(Theta_lower_f_index:Theta_upper_f_index);
f_Alpha=f(Alpha_lower_f_index:Alpha_upper_f_index);
f_Beta=f(Beta_lower_f_index:Beta_upper_f_index);
f_Gamma=f(Gamma_lower_f_index:Gamma_upper_f_index);

% [Tf,~,~]=multi_sst_TF_main_stft_based_v2(IMF,5,1,fs,fr,W_STFT,alpha);
% 
% Cxy_Delta=single(zeros(NUM_IMFs,NUM_PAIRS,Delta_upper_f_index-Delta_lower_f_index+1,size(Tf,2)));
% Cxy_Theta=single(zeros(NUM_IMFs,NUM_PAIRS,Theta_upper_f_index-Theta_lower_f_index+1,size(IMF,3)));
% Cxy_Alpha=single(zeros(NUM_IMFs,NUM_PAIRS,Alpha_upper_f_index-Alpha_lower_f_index+1,size(IMF,3)));
% Cxy_Beta=single(zeros(NUM_IMFs,NUM_PAIRS,Beta_upper_f_index-Beta_lower_f_index+1,size(IMF,3)));
% Cxy_Gamma=single(zeros(NUM_IMFs,NUM_PAIRS,Gamma_upper_f_index-Gamma_lower_f_index+1,size(Tf,2)));
% Tf=[];
% x=[];
% sst=[];
for pair_count=1:NUM_PAIRS
    for i=1:NUM_IMFs
        x=[];
        x(:,1)=squeeze(IMF(pairs(pair_count,1),i,:));
        x(:,2)=squeeze(IMF(pairs(pair_count,2),i,:));
                 
        [Tf,~,sst]=multi_sst_TF_main_stft_based_v2(x,5,0,fs,fr,W_STFT,alpha); % 1 is worse, more ripples
        
        S12=single(sqrt(Tf));
        
        Tf=[];

        S1=single(abs(squeeze(sst(1,:,:))));
        S2=single(abs(squeeze(sst(2,:,:))));
                
        sst=[];
        
        S_1_2=single(sqrt(S1.*S2));
        
        S1=[];
        S2=[];

        Cxy_temp = single(S_1_2./S12);

        S_1_2=[];
        S12=[];
        
        Cxy_temp(find(Cxy_temp==Inf))=0;
        Cxy_temp(isnan(Cxy_temp))=0;

        for band_count=1:5
            if band_count==1
                Cxy_Delta_temp=Cxy_temp(Delta_lower_f_index:Delta_upper_f_index,:);      
            elseif band_count==2
                Cxy_Theta_temp=Cxy_temp(Theta_lower_f_index:Theta_upper_f_index,:);
            elseif band_count==3
                Cxy_Alpha_temp=Cxy_temp(Alpha_lower_f_index:Alpha_upper_f_index,:);
            elseif band_count==4
                Cxy_Beta_temp=Cxy_temp(Beta_lower_f_index:Beta_upper_f_index,:);
            elseif band_count==5
                Cxy_Gamma_temp=Cxy_temp(Gamma_lower_f_index:Gamma_upper_f_index,:);
            end
                        
        end
        
        Cxy_temp=[];
        
         for band_count=1:5
            if band_count==1   
                Cxy_Delta(i,pair_count,:,:) =ISC_sub(Cxy_Delta_temp,Th_std);      
                Cxy_Delta_temp=[];
            elseif band_count==2
                Cxy_Theta(i,pair_count,:,:) =ISC_sub(Cxy_Theta_temp,Th_std);
                Cxy_Theta_temp=[];
            elseif band_count==3
                Cxy_Alpha(i,pair_count,:,:) =ISC_sub(Cxy_Alpha_temp,Th_std);
                Cxy_Alpha_temp=[];
            elseif band_count==4
                Cxy_Beta(i,pair_count,:,:) =ISC_sub(Cxy_Beta_temp,Th_std);
                Cxy_Beta_temp=[];
            elseif band_count==5
                Cxy_Gamma(i,pair_count,:,:) =ISC_sub(Cxy_Gamma_temp,Th_std);
                Cxy_Gamma_temp=[];
            end
                        
         end
    
          
    end
end
 
f_Delta=linspace(Delta_lower_f,Delta_upper_f,size(Cxy_Delta,4));
f_Theta=linspace(Theta_lower_f,Theta_upper_f,size(Cxy_Theta,4));
f_Alpha=linspace(Alpha_lower_f,Alpha_upper_f,size(Cxy_Alpha,4));
f_Beta=linspace(Beta_lower_f,Beta_upper_f,size(Cxy_Beta,4));
f_Gamma=linspace(Gamma_lower_f,Gamma_upper_f,size(Cxy_Gamma,4));
% --------------END OF ISC_STFT_EEG --------------

function [Cxy_band]=ISC_sub(Cxy_band_temp,Th_std)
        
        Band_max=max(Cxy_band_temp);

        Band_max_std=std(Band_max);

        I_temp_band=find(Cxy_band_temp>Th_std*Band_max_std);

        Cxy_band_temp(I_temp_band)=Th_std*Band_max_std;

        Cxy_band=single(Cxy_band_temp/max(max(Cxy_band_temp)));

